#include "reduction.cuh"

#include <cuda_runtime.h>
#include <algorithm>

// Reduction in the global memory
__global__ void reduceKernel_v0(unsigned int *g_idata, unsigned int *g_odata, size_t dataSize)
{
	// load shared memory
	unsigned int lid = threadIdx.x; // local id in the block
	unsigned int id = blockIdx.x*blockDim.x + threadIdx.x; // global id, base index of the thread block

	if (id >= dataSize) return;

	// do reduction in global memory
	for (unsigned int s = 1; s < blockDim.x; s *= 2)
	{
		// modulo arithmetic is slow!
		if ((lid % (2 * s)) == 0)
		{
			if (id + s < dataSize)	g_idata[id] += g_idata[id + s];
		}

		__syncthreads();
	}

	// write result for this block to global memory
	if (lid == 0) g_odata[blockIdx.x] = g_idata[id];
}

// Reduction in shared memory
__global__ void reduceKernel_v1(unsigned int *g_idata, unsigned int *g_odata, size_t dataSize)
{
	__shared__ unsigned int sdata[threadsPerBlock];

	// load shared memory
	unsigned int lid = threadIdx.x; // local id in the block
	unsigned int id = blockIdx.x*blockDim.x + threadIdx.x; // global id, index of the data

	sdata[lid] = (id < dataSize) ? g_idata[id] : 0;

	__syncthreads();

	// do reduction in shared memory
	for (unsigned int s = 1; s < blockDim.x; s *= 2)
	{
		// modulo arithmetic is slow!
		if ((lid % (2 * s)) == 0)
		{
			sdata[lid] += sdata[lid + s];
		}

		__syncthreads();
	}

	// write result for this block to global memory
	if (lid == 0) g_odata[blockIdx.x] = sdata[0];
}

__global__ void reduceKernel_v2(unsigned int *g_idata, unsigned int *g_odata, size_t dataSize)
{
	__shared__ unsigned int sdata[threadsPerBlock];

	// load shared memory
	unsigned int lid = threadIdx.x; // local id in the block
	unsigned int id = blockIdx.x*blockDim.x + threadIdx.x; // global id, index of the data

	sdata[lid] = (id < dataSize) ? g_idata[id] : 0;

	__syncthreads();

	// do reduction in shared memory
	for (unsigned int s = 1; s < blockDim.x; s *= 2)
	{
		int index = 2 * s * lid;
		if (index < blockDim.x)
		{
			sdata[index] += sdata[index + s];
		}

		__syncthreads();
	}

	// write result for this block to global memory
	if (lid == 0) g_odata[blockIdx.x] = sdata[0];
}

__global__ void reduceKernel_v3(unsigned int *g_idata, unsigned int *g_odata, size_t dataSize)
{
	__shared__ unsigned int sdata[threadsPerBlock];

	// load shared memory
	unsigned int lid = threadIdx.x; // local id in the block
	unsigned int id = blockIdx.x*blockDim.x + threadIdx.x; // global id, index of the data

	sdata[lid] = (id < dataSize) ? g_idata[id] : 0;

	__syncthreads();

	// do reduction in shared memory
	for (unsigned int s = blockDim.x / 2; s > 0; s >>= 1)
	{
		if (lid < s)
		{
			sdata[lid] += sdata[lid + s];
		}

		__syncthreads();
	}

	// write result for this block to global memory
	if (lid == 0) g_odata[blockIdx.x] = sdata[0];
}

__global__ void reduceKernel_v4(unsigned int *g_idata, unsigned int *g_odata, size_t dataSize)
{
	__shared__ unsigned int sdata[threadsPerBlock];

	// perform first level of reduction,
	// reading from global memory, writing to shared memory
	unsigned int lid = threadIdx.x;
	unsigned int id = blockIdx.x*(blockDim.x * 2) + threadIdx.x;

	unsigned int mySum = (id < dataSize) ? g_idata[id] : 0;

	if (id + blockDim.x < dataSize)
		mySum += g_idata[id + blockDim.x];

	sdata[lid] = mySum;
	__syncthreads();

	// do reduction in shared memory
	for (unsigned int s = blockDim.x / 2; s>0; s >>= 1)
	{
		if (lid < s)
		{
			sdata[lid] = mySum = mySum + sdata[lid + s];
		}

		__syncthreads();
	}

	// write result for this block to global memory
	if (lid == 0) g_odata[blockIdx.x] = mySum;
}

template<unsigned int blockSize>
__global__ void reduceKernel_v5(unsigned int *g_idata, unsigned *g_odata, size_t dataSize)
{
	__shared__ unsigned int sdata[threadsPerBlock];

	// perform first level of reduction,
	// reading from global memory, writing to shared memory
	unsigned int lid = threadIdx.x;
	unsigned int id = blockIdx.x*(blockDim.x * 2) + threadIdx.x;

	unsigned int mySum = (id < dataSize) ? g_idata[id] : 0;

	if (id + blockSize < dataSize)
	{
		mySum += g_idata[id + blockSize];
	}

	sdata[lid] = mySum;
	__syncthreads();

	// do reduction in shared memory down to warp size
	for (unsigned int s = blockDim.x / 2; s > 32; s >>= 1)
	{
		if (lid < s)
		{
			sdata[lid] = mySum = mySum + sdata[lid + s];
		}
		__syncthreads();
	}

#if (__CUDA_ARCH__ >= 300 )
	if (lid < 32)
	{
		// Fetch final intermediate sum from 2nd warp
		if (blockSize >= 64) mySum += sdata[lid + 32];
		// Reduce final warp using shuffle
		for (int offset = warpSize / 2; offset > 0; offset /= 2)
		{
			mySum += __shfl_down(mySum, offset);
		}
	}
#else
	// fully unroll reduction within a single warp
	if ((blockSize >= 64) && (lid < 32))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 32];
	}
	__syncthreads();

	if ((blockSize >= 32) && (lid < 16))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 16];
	}
	__syncthreads();

	if ((blockSize >= 16) && (lid <  8))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 8];
	}
	__syncthreads();

	if ((blockSize >= 8) && (lid <  4))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 4];
	}
	__syncthreads();

	if ((blockSize >= 4) && (lid <  2))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 2];
	}
	__syncthreads();

	if ((blockSize >= 2) && (lid <  1))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 1];
	}
	__syncthreads();
#endif

	// write result for this block to global memory
	if (lid == 0) g_odata[blockIdx.x] = mySum;
}

template<unsigned int blockSize>
__global__ void reduceKernel_v6(unsigned int *g_idata, unsigned *g_odata, size_t dataSize)
{
	__shared__ unsigned int sdata[threadsPerBlock];

	// perform first level of reduction,
	// reading from global memory, writing to shared memory
	unsigned int lid = threadIdx.x;
	unsigned int id = blockIdx.x*(blockDim.x * 2) + threadIdx.x;

	unsigned int mySum = (id < dataSize) ? g_idata[id] : 0;

	if (id + blockSize < dataSize)
	{
		mySum += g_idata[id + blockSize];
	}

	sdata[lid] = mySum;
	__syncthreads();

	// do fully unrolled reduction in shared memory down to warp size
	if ((blockSize >= 512) && (lid < 256))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 256];
	}
	__syncthreads();

	if ((blockSize >= 256) && (lid < 128))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 128];
	}
	__syncthreads();

	if ((blockSize >= 128) && (lid <  64))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 64];
	}
	__syncthreads();

#if (__CUDA_ARCH__ >= 300 )
	if (lid < 32)
	{
		// Fetch final intermediate sum from 2nd warp
		if (blockSize >= 64) mySum += sdata[lid + 32];
		// Reduce final warp using shuffle
		for (int offset = warpSize / 2; offset > 0; offset /= 2)
		{
			mySum += __shfl_down(mySum, offset);
		}
	}
#else
	// fully unroll reduction within a single warp
	if ((blockSize >= 64) && (lid < 32))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 32];
	}
	__syncthreads();

	if ((blockSize >= 32) && (lid < 16))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 16];
	}
	__syncthreads();

	if ((blockSize >= 16) && (lid <  8))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 8];
	}
	__syncthreads();

	if ((blockSize >= 8) && (lid <  4))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 4];
	}
	__syncthreads();

	if ((blockSize >= 4) && (lid <  2))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 2];
	}
	__syncthreads();

	if ((blockSize >= 2) && (lid <  1))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 1];
	}
	__syncthreads();
#endif

	// write result for this block to global mem
	if (lid == 0) g_odata[blockIdx.x] = mySum;
}

template<unsigned int blockSize>
__global__ void reduceKernel_v7(unsigned int *g_idata, unsigned *g_odata, size_t dataSize)
{
	__shared__ unsigned int sdata[threadsPerBlock];

	// perform first level of reduction,
	// reading from global memory, writing to shared memory
	unsigned int lid = threadIdx.x;
	unsigned int id = blockIdx.x*(blockDim.x * 2) + threadIdx.x;
	unsigned int gridSize = blockSize * 2 * gridDim.x;

	unsigned int mySum = 0;

	// we reduce multiple elements per thread.  The number is determined by the
	// number of active thread blocks (via gridDim).  More blocks will result
	// in a larger gridSize and therefore fewer elements per thread
	while (id < dataSize)
	{
		mySum += g_idata[id];

		// ensure we don't read out of bounds
		if (id + blockSize < dataSize)
		{
			mySum += g_idata[id + blockSize];
		}
		id += gridSize;
	}

	sdata[lid] = mySum;
	__syncthreads();

	// do fully unrolled reduction in shared memory down to warp size
	if ((blockSize >= 512) && (lid < 256))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 256];
	}
	__syncthreads();

	if ((blockSize >= 256) && (lid < 128))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 128];
	}
	__syncthreads();

	if ((blockSize >= 128) && (lid <  64))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 64];
	}
	__syncthreads();

#if (__CUDA_ARCH__ >= 300 )
	if (lid < 32)
	{
		// Fetch final intermediate sum from 2nd warp
		if (blockSize >= 64) mySum += sdata[lid + 32];
		// Reduce final warp using shuffle
		for (int offset = warpSize / 2; offset > 0; offset /= 2)
		{
			mySum += __shfl_down(mySum, offset);
		}
	}
#else
	// fully unroll reduction within a single warp
	if ((blockSize >= 64) && (lid < 32))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 32];
	}
	__syncthreads();

	if ((blockSize >= 32) && (lid < 16))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 16];
	}
	__syncthreads();

	if ((blockSize >= 16) && (lid <  8))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 8];
	}
	__syncthreads();

	if ((blockSize >= 8) && (lid <  4))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 4];
	}
	__syncthreads();

	if ((blockSize >= 4) && (lid <  2))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 2];
	}
	__syncthreads();

	if ((blockSize >= 2) && (lid <  1))
	{
		sdata[lid] = mySum = mySum + sdata[lid + 1];
	}
	__syncthreads();
#endif

	// write result for this block to global mem
	if (lid == 0) g_odata[blockIdx.x] = mySum;
}

void reduceGPU_v0(unsigned int* d_idata, unsigned int* d_odata, size_t dataSize)
{
	dim3 blockDim(threadsPerBlock, 1, 1);
	// Number of blocks: ceil(dataSize / threadsPerBlock). Cast from size_t to uint to eliminate warning.
	dim3 gridDim(static_cast<unsigned int>((dataSize + threadsPerBlock - 1) / threadsPerBlock), 1, 1);

	// Compute partial reduction. Each block writes its partial sum to d_odata.
	reduceKernel_v0<<<gridDim, blockDim>>>(d_idata, d_odata, dataSize);
	
	// Continue the reduction on the partial sums
	dataSize = gridDim.x;
	while (dataSize > 1)
	{
		// Copy the output buffer to the input buffer (alternative: ping-pong)
		cudaMemcpy(d_idata, d_odata, dataSize * sizeof(unsigned int), cudaMemcpyDeviceToDevice);

		gridDim = dim3(static_cast<unsigned int>((dataSize + threadsPerBlock - 1) / threadsPerBlock), 1, 1);
		blockDim.x = static_cast<unsigned int>(std::min(static_cast<size_t>(blockDim.x), dataSize));
		reduceKernel_v0<<<gridDim, blockDim>>>(d_idata, d_odata, dataSize);
		dataSize = gridDim.x;
	}
}

void reduceGPU_v1(unsigned int* d_idata, unsigned int* d_odata, size_t dataSize)
{
	dim3 blockDim(threadsPerBlock, 1, 1);
	// Number of blocks: ceil(dataSize / threadsPerBlock). Cast from size_t to uint to eliminate warning.
	dim3 gridDim(static_cast<unsigned int>((dataSize + threadsPerBlock - 1) / threadsPerBlock), 1, 1);

	// Compute partial reduction. Each block writes its partial sum to d_odata.
	reduceKernel_v1<<<gridDim, blockDim>>>(d_idata, d_odata, dataSize);
	
	// Continue the reduction on the partial sums
	dataSize = gridDim.x;
	while (dataSize > 1)
	{
		// Copy the output buffer to the input buffer (alternative: ping-pong)
		cudaMemcpy(d_idata, d_odata, dataSize * sizeof(unsigned int), cudaMemcpyDeviceToDevice);

		gridDim = dim3(static_cast<unsigned int>((dataSize + threadsPerBlock - 1) / threadsPerBlock), 1, 1);
		blockDim.x = static_cast<unsigned int>(std::min(static_cast<size_t>(blockDim.x), dataSize));
		reduceKernel_v1<<<gridDim, blockDim>>>(d_idata, d_odata, dataSize);
		dataSize = gridDim.x;
	}
}

void reduceGPU_v2(unsigned int* d_idata, unsigned int* d_odata, size_t dataSize)
{
	dim3 blockDim(threadsPerBlock, 1, 1);

	// Same as v0 with everything merged in a single loop
	while (dataSize > 1)
	{
		// Number of blocks: ceil(dataSize / threadsPerBlock). Cast from size_t to uint to eliminate warning.
		dim3 gridDim(static_cast<unsigned int>((dataSize + threadsPerBlock - 1) / threadsPerBlock), 1, 1);
		blockDim.x = static_cast<unsigned int>(std::min(static_cast<size_t>(blockDim.x), dataSize));
		reduceKernel_v2<<<gridDim, blockDim>>>(d_idata, d_odata, dataSize);

		dataSize = gridDim.x;
		// Copy the output buffer to the input buffer (alternative: ping-pong)
		if (dataSize > 1)
		{
			cudaMemcpy(d_idata, d_odata, dataSize * sizeof(unsigned int), cudaMemcpyDeviceToDevice);
		}
	}
}

void reduceGPU_v3(unsigned int* d_idata, unsigned int* d_odata, size_t dataSize)
{
	dim3 blockDim(threadsPerBlock, 1, 1);

	// Same as v1 but with ping-pong instead of copying
	bool resultInOdata = false;
	unsigned int* pIn = d_idata;
	unsigned int* pOut = d_odata;
	while (dataSize > 1)
	{
		// Number of blocks: ceil(dataSize / threadsPerBlock). Cast from size_t to uint to eliminate warning.
		dim3 gridDim(static_cast<unsigned int>((dataSize + threadsPerBlock - 1) / threadsPerBlock), 1, 1);
		blockDim.x = static_cast<unsigned int>(std::min(static_cast<size_t>(blockDim.x), dataSize));
		reduceKernel_v3<<<gridDim, blockDim>>>(pIn, pOut, dataSize);

		dataSize = gridDim.x;
		std::swap(pIn, pOut); // Swaps only the pointers!
		resultInOdata = !resultInOdata;
	}
	// The host code expects the result in d_odata so we need to put it there
	if (!resultInOdata)
	{
		cudaMemcpy(d_odata, d_idata, dataSize * sizeof(unsigned int), cudaMemcpyDeviceToDevice);
	}
}

void reduceGPU_v4(unsigned int* d_idata, unsigned int* d_odata, size_t dataSize)
{
	dim3 blockDim(threadsPerBlock, 1, 1);

	// Same as v1 but with ping-pong instead of copying
	bool resultInOdata = false;
	unsigned int* pIn = d_idata;
	unsigned int* pOut = d_odata;
	while (dataSize > 1)
	{
		// Number of blocks: ceil(dataSize / threadsPerBlock*2). Cast from size_t to uint to eliminate warning.
		dim3 gridDim(static_cast<unsigned int>((dataSize + threadsPerBlock*2 - 1) / (threadsPerBlock*2)), 1, 1);
		blockDim.x = static_cast<unsigned int>(std::min(static_cast<size_t>(blockDim.x), dataSize));
		reduceKernel_v4<<<gridDim, blockDim>>>(pIn, pOut, dataSize);

		dataSize = gridDim.x;
		std::swap(pIn, pOut); // Swaps only the pointers!
		resultInOdata = !resultInOdata;
	}
	// The host code expects the result in d_odata so we need to put it there
	if (!resultInOdata)
	{
		cudaMemcpy(d_odata, d_idata, dataSize * sizeof(unsigned int), cudaMemcpyDeviceToDevice);
	}
}

void reduceGPU_v5(unsigned int* d_idata, unsigned int* d_odata, size_t dataSize)
{
	dim3 blockDim(threadsPerBlock, 1, 1);

	// Same as v1 but with ping-pong instead of copying
	bool resultInOdata = false;
	unsigned int* pIn = d_idata;
	unsigned int* pOut = d_odata;
	while (dataSize > 1)
	{
		// Number of blocks: ceil(dataSize / threadsPerBlock*2). Cast from size_t to uint to eliminate warning.
		dim3 gridDim(static_cast<unsigned int>((dataSize + threadsPerBlock*2 - 1) / (threadsPerBlock*2)), 1, 1);
		blockDim.x = static_cast<unsigned int>(std::min(static_cast<size_t>(blockDim.x), dataSize));
		switch(blockDim.x)
		{
		case 1024:
			reduceKernel_v5<1024><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 512:
			reduceKernel_v5<512><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 256:
			reduceKernel_v5<256><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 128:
			reduceKernel_v5<128><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 64:
			reduceKernel_v5<64><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 32:
			reduceKernel_v5<32><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 16:
			reduceKernel_v5<16><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 8:
			reduceKernel_v5<8><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 4:
			reduceKernel_v5<4><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 2:
			reduceKernel_v5<2><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 1:
			reduceKernel_v5<1><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		}

		dataSize = gridDim.x;
		std::swap(pIn, pOut); // Swaps only the pointers!
		resultInOdata = !resultInOdata;
	}
	// The host code expects the result in d_odata so we need to put it there
	if (!resultInOdata)
	{
		cudaMemcpy(d_odata, d_idata, dataSize * sizeof(unsigned int), cudaMemcpyDeviceToDevice);
	}
}

void reduceGPU_v6(unsigned int* d_idata, unsigned int* d_odata, size_t dataSize)
{
	dim3 blockDim(threadsPerBlock, 1, 1);

	// Same as v1 but with ping-pong instead of copying
	bool resultInOdata = false;
	unsigned int* pIn = d_idata;
	unsigned int* pOut = d_odata;
	while (dataSize > 1)
	{
		// Number of blocks: ceil(dataSize / threadsPerBlock*2). Cast from size_t to uint to eliminate warning.
		dim3 gridDim(static_cast<unsigned int>((dataSize + threadsPerBlock * 2 - 1) / (threadsPerBlock * 2)), 1, 1);
		blockDim.x = static_cast<unsigned int>(std::min(static_cast<size_t>(blockDim.x), dataSize));
		switch (blockDim.x)
		{
		case 1024:
			reduceKernel_v6<1024><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 512:
			reduceKernel_v6<512><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 256:
			reduceKernel_v6<256><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 128:
			reduceKernel_v6<128><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 64:
			reduceKernel_v6<64><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 32:
			reduceKernel_v6<32><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 16:
			reduceKernel_v6<16><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 8:
			reduceKernel_v6<8><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 4:
			reduceKernel_v6<4><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 2:
			reduceKernel_v6<2><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 1:
			reduceKernel_v6<1><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		}

		dataSize = gridDim.x;
		std::swap(pIn, pOut); // Swaps only the pointers!
		resultInOdata = !resultInOdata;
	}
	// The host code expects the result in d_odata so we need to put it there
	if (!resultInOdata)
	{
		cudaMemcpy(d_odata, d_idata, dataSize * sizeof(unsigned int), cudaMemcpyDeviceToDevice);
	}
}

void reduceGPU_v7(unsigned int* d_idata, unsigned int* d_odata, size_t dataSize)
{
	dim3 blockDim(threadsPerBlock, 1, 1);

	// Same as v1 but with ping-pong instead of copying
	bool resultInOdata = false;
	unsigned int* pIn = d_idata;
	unsigned int* pOut = d_odata;
	while (dataSize > 1)
	{
		// Number of blocks: ceil(dataSize / threadsPerBlock*2). Cast from size_t to uint to eliminate warning.
		dim3 gridDim(static_cast<unsigned int>((dataSize + threadsPerBlock * 2 - 1) / (threadsPerBlock * 2)), 1, 1);
		blockDim.x = static_cast<unsigned int>(std::min(static_cast<size_t>(blockDim.x), dataSize));
		switch (blockDim.x)
		{
		case 1024:
			reduceKernel_v7<1024><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 512:
			reduceKernel_v7<512><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 256:
			reduceKernel_v7<256><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 128:
			reduceKernel_v7<128><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 64:
			reduceKernel_v7<64><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 32:
			reduceKernel_v7<32><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 16:
			reduceKernel_v7<16><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 8:
			reduceKernel_v7<8><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 4:
			reduceKernel_v7<4><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 2:
			reduceKernel_v7<2><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		case 1:
			reduceKernel_v7<1><<<gridDim, blockDim>>>(pIn, pOut, dataSize);
			break;
		}

		dataSize = gridDim.x;
		std::swap(pIn, pOut); // Swaps only the pointers!
		resultInOdata = !resultInOdata;
	}
	// The host code expects the result in d_odata so we need to put it there
	if (!resultInOdata)
	{
		cudaMemcpy(d_odata, d_idata, dataSize * sizeof(unsigned int), cudaMemcpyDeviceToDevice);
	}
}

void reduceGPU(unsigned int* d_idata, unsigned int* d_odata, size_t dataSize, int method, bool synchronizeDevice)
{
	switch (method)
	{
	case 0:
		reduceGPU_v0(d_idata, d_odata, dataSize);
		break;
	case 1:
		reduceGPU_v1(d_idata, d_odata, dataSize);
		break;
	case 2:
		reduceGPU_v2(d_idata, d_odata, dataSize);
		break;
	case 3:
		reduceGPU_v3(d_idata, d_odata, dataSize);
		break;
	case 4:
		reduceGPU_v4(d_idata, d_odata, dataSize);
		break;
	case 5:
		reduceGPU_v5(d_idata, d_odata, dataSize);
		break;
	case 6:
		reduceGPU_v6(d_idata, d_odata, dataSize);
		break;
	case 7:
		reduceGPU_v7(d_idata, d_odata, dataSize);
		break;
	}


	if (synchronizeDevice) cudaDeviceSynchronize();
}